from dataclasses import dataclass

from leeger.enum.MatchupType import MatchupType
from leeger.model.abstract.UniqueId import UniqueId
from leeger.model.league.Matchup import Matchup


@dataclass(kw_only=True)
class Week(UniqueId):
    weekNumber: int
    matchups: list[Matchup]

    @property
    def isPlayoffWeek(self) -> bool:
        isPlayoffWeek = False
        for matchup in self.matchups:
            if matchup.matchupType == MatchupType.PLAYOFF:
                isPlayoffWeek = True
                break
        return isPlayoffWeek or self.isChampionshipWeek

    @property
    def isChampionshipWeek(self) -> bool:
        isChampionshipWeek = False
        for matchup in self.matchups:
            if matchup.matchupType == MatchupType.CHAMPIONSHIP:
                isChampionshipWeek = True
                break
        return isChampionshipWeek



import unittest

from espn_api.requests.espn_requests import ESPNInvalidLeague

from leeger.league_loader.ESPNLeagueLoader import ESPNLeagueLoader


class TestESPNLeagueLoader(unittest.TestCase):
    """
    Currently, there is not a good way to test this class due to an issue addressed here: https://github.com/cwendt94/espn-api/issues/338
    For now, we can do some very basic tests.
    """

    def test_loadLeague_intendedFailure(self):
        with self.assertRaises(ESPNInvalidLeague) as context:
            ESPNLeagueLoader.loadLeague(0, [2000])  # 0 is a bad league ID
        self.assertEqual("League 0 does not exist", str(context.exception))



import unittest

from leeger.util.GeneralUtil import GeneralUtil


class TestGeneralUtil(unittest.TestCase):
    def test_filter_None(self):
        l = [None, "test", None]
        response = GeneralUtil.filter(value=None, list_=l)
        self.assertEqual(["test"], response)

    def test_filter_emptyList(self):
        l = list()
        response = GeneralUtil.filter(value=None, list_=l)
        self.assertEqual([], response)



import unittest

from leeger.model.league.Owner import Owner


class TestOwner(unittest.TestCase):
    def test_owner_init(self):
        owner = Owner(name="name")

        self.assertEqual("name", owner.name)



import unittest

from leeger.league_loader.YahooLeagueLoader import YahooLeagueLoader


class TestESPNLeagueLoader(unittest.TestCase):
    """
    # TODO: Add better tests.
    """

    def test_loadLeague_intendedFailure(self):
        with self.assertRaises(TimeoutError) as context:
            badClientId = badClientSecret = "bad"
            badLeagueId = 0
            YahooLeagueLoader.loadLeague(badLeagueId, [2000], clientId=badClientId,
                                         clientSecret=badClientSecret, loginTimeoutSeconds=1)
        self.assertEqual("Login to yahoofantasy timed out.", str(context.exception))



import unittest

from leeger.model.league.Team import Team


class TestTeam(unittest.TestCase):
    def test_team_init(self):
        team = Team(ownerId="ownerId", name="name")

        self.assertEqual("ownerId", team.ownerId)
        self.assertEqual("name", team.name)



from leeger.calculator.parent.StreakCalculator import StreakCalculator
from leeger.decorator.validate.validators import validateLeague
from leeger.model.league.League import League
from leeger.model.league.Team import Team
from leeger.model.stat.Streak import Streak
from leeger.util.LeagueNavigator import LeagueNavigator
from leeger.util.MatchupNavigator import MatchupNavigator


class GameOutcomeStreakCalculator(StreakCalculator):
    """
    Used to calculate all Streaks for Game Outcomes.
    """

    @classmethod
    @validateLeague
    def getWinStreaks(cls, league: League, **kwargs) -> list[Streak]:
        """
        Returns all win streaks in the given League.
        """
        streakFilters = cls._getStreakFilters(league, **kwargs)
        allFilteredMatchupsByWeek = cls._getAllFilteredMatchupsByWeek(league, streakFilters, validateLeague=False)
        weekNumberAndTeamsThatWon: dict[int, list[Team]] = dict()
        for weekNumber in allFilteredMatchupsByWeek.keys():
            weekNumberAndTeamsThatWon[weekNumber] = list()
            matchups = allFilteredMatchupsByWeek[weekNumber]
            for matchup in matchups:
                teamIdOfMatchupWinner = MatchupNavigator.getTeamIdOfMatchupWinner(matchup, validateMatchup=False)
                if teamIdOfMatchupWinner is not None:
                    team = LeagueNavigator.getTeamById(teamIdOfMatchupWinner)
                    weekNumberAndTeamsThatWon[weekNumber].append(team)

        currentStreaks: list[Streak] = list()
        for weekNumber in weekNumberAndTeamsThatWon.keys():
            for winningTeam in weekNumberAndTeamsThatWon[weekNumber]:
                ...

    @classmethod
    def __ownerIsOnStreak(cls, league: League, team: Team, currentWeekNumber: int,
                          currentStreaks: list[Streak]) -> bool:
        """
        Returns whether the Owner of the given Team is on a streak or not based on the list of current Streaks.
        """
        streaksWithThisOwner = [streak for streak in currentStreaks if
                                LeagueNavigator.getTeamById(league, streak.teamIdStart).ownerId == team.ownerId]
        return



import os
import random
from dataclasses import dataclass

from openpyxl import Workbook, load_workbook
from openpyxl.styles import Font, Color, PatternFill
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.table import Table

from leeger.model.abstract.UniqueId import UniqueId
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.stat.YearStatSheet import YearStatSheet


@dataclass(kw_only=True)
class Year(UniqueId):
    yearNumber: int
    teams: list[Team]
    weeks: list[Week]

    def statSheet(self, **kwargs) -> YearStatSheet:
        from leeger.calculator.year_calculator.AWALYearCalculator import AWALYearCalculator
        from leeger.calculator.year_calculator.GameOutcomeYearCalculator import GameOutcomeYearCalculator
        from leeger.calculator.year_calculator.PlusMinusYearCalculator import PlusMinusYearCalculator
        from leeger.calculator.year_calculator.PointsScoredYearCalculator import PointsScoredYearCalculator
        from leeger.calculator.year_calculator.ScoringShareYearCalculator import ScoringShareYearCalculator
        from leeger.calculator.year_calculator.ScoringStandardDeviationYearCalculator import \
            ScoringStandardDeviationYearCalculator
        from leeger.calculator.year_calculator.SingleScoreYearCalculator import SingleScoreYearCalculator
        from leeger.calculator.year_calculator.SmartWinsYearCalculator import SmartWinsYearCalculator
        from leeger.calculator.year_calculator.SSLYearCalculator import SSLYearCalculator

        # Game Outcome
        wins = GameOutcomeYearCalculator.getWins(self, **kwargs)
        losses = GameOutcomeYearCalculator.getLosses(self, **kwargs)
        ties = GameOutcomeYearCalculator.getTies(self, **kwargs)
        winPercentage = GameOutcomeYearCalculator.getWinPercentage(self, **kwargs)
        wal = GameOutcomeYearCalculator.getWAL(self, **kwargs)
        walPerGame = GameOutcomeYearCalculator.getWALPerGame(self, **kwargs)

        # AWAL
        awal = AWALYearCalculator.getAWAL(self, **kwargs)
        awalPerGame = AWALYearCalculator.getAWALPerGame(self, **kwargs)
        opponentAWAL = AWALYearCalculator.getOpponentAWAL(self, **kwargs)
        opponentAWALPerGame = AWALYearCalculator.getOpponentAWALPerGame(self, **kwargs)

        # Smart Wins
        smartWins = SmartWinsYearCalculator.getSmartWins(self, **kwargs)
        smartWinsPerGame = SmartWinsYearCalculator.getSmartWinsPerGame(self, **kwargs)
        opponentSmartWins = SmartWinsYearCalculator.getOpponentSmartWins(self, **kwargs)
        opponentSmartWinsPerGame = SmartWinsYearCalculator.getOpponentSmartWinsPerGame(self, **kwargs)

        # Points Scored
        pointsScored = PointsScoredYearCalculator.getPointsScored(self, **kwargs)
        pointsScoredPerGame = PointsScoredYearCalculator.getPointsScoredPerGame(self, **kwargs)
        opponentPointsScored = PointsScoredYearCalculator.getOpponentPointsScored(self, **kwargs)
        opponentPointsScoredPerGame = PointsScoredYearCalculator.getOpponentPointsScoredPerGame(self, **kwargs)

        # Scoring Share
        scoringShare = ScoringShareYearCalculator.getScoringShare(self, **kwargs)
        opponentScoringShare = ScoringShareYearCalculator.getOpponentScoringShare(self, **kwargs)

        # Single Score
        maxScore = SingleScoreYearCalculator.getMaxScore(self, **kwargs)
        minScore = SingleScoreYearCalculator.getMinScore(self, **kwargs)

        # Scoring Standard Deviation
        scoringStandardDeviation = ScoringStandardDeviationYearCalculator.getScoringStandardDeviation(self, **kwargs)

        # Plus Minus
        plusMinus = PlusMinusYearCalculator.getPlusMinus(self, **kwargs)

        # SSL
        teamScore = SSLYearCalculator.getTeamScore(self, **kwargs)
        teamSuccess = SSLYearCalculator.getTeamSuccess(self, **kwargs)
        teamLuck = SSLYearCalculator.getTeamLuck(self, **kwargs)

        return YearStatSheet(wins=wins, losses=losses, ties=ties, winPercentage=winPercentage, wal=wal,
                             walPerGame=walPerGame, awal=awal, awalPerGame=awalPerGame, opponentAWAL=opponentAWAL,
                             opponentAWALPerGame=opponentAWALPerGame, smartWins=smartWins,
                             smartWinsPerGame=smartWinsPerGame, opponentSmartWins=opponentSmartWins,
                             opponentSmartWinsPerGame=opponentSmartWinsPerGame, pointsScored=pointsScored,
                             pointsScoredPerGame=pointsScoredPerGame, opponentPointsScored=opponentPointsScored,
                             opponentPointsScoredPerGame=opponentPointsScoredPerGame, scoringShare=scoringShare,
                             opponentScoringShare=opponentScoringShare, maxScore=maxScore, minScore=minScore,
                             scoringStandardDeviation=scoringStandardDeviation, plusMinus=plusMinus,
                             teamScore=teamScore, teamSuccess=teamSuccess, teamLuck=teamLuck)

    def toExcel(self, filePath: str, **kwargs) -> None:
        """
        If the given Excel file exists already, add a worksheet to it with this year.
        If the given Excel file does not exist, create a new workbook and add a worksheet to it with this year.
        """

        if os.path.exists(filePath):
            workbook = load_workbook(filename=filePath)
            # figure out index to put this sheet into
            # we want the years as sheets in order from oldest -> newest year
            index = 0
            for i, sheetname in enumerate(workbook.sheetnames):
                if self.yearNumber > int(sheetname):
                    index += 1
            workbook.create_sheet(str(self.yearNumber), index=index)
        else:
            # create workbook and sheet
            workbook = Workbook()
            workbook.create_sheet(str(self.yearNumber), index=0)
            # remove default sheet
            del workbook["Sheet"]
        worksheet = workbook[str(self.yearNumber)]

        ####################
        # Styles for table #
        ####################

        # fonts
        headerColumnFont = Font(size=12, bold=True)
        teamNameFont = Font(size=11, bold=True)

        # colors
        GRAY = Color(rgb="B8B8B8")

        def getRandomColor(tint: float = 0) -> Color:
            r = lambda: random.randint(0, 255)
            hexCode = "%02X%02X%02X" % (r(), r(), r())
            return Color(rgb=hexCode, tint=tint)

        TEAM_ROW_COLORS = [getRandomColor(0.5) for _ in range(len(self.teams))]

        # fills
        headerFill = PatternFill(patternType="solid", fgColor=GRAY)

        #################
        # Fill in table #
        #################

        # add title
        worksheet["A1"] = "Team Names"
        worksheet["A1"].font = headerColumnFont
        worksheet["A1"].fill = headerFill
        # add all team names
        for i, team in enumerate(self.teams):
            col = "A"
            worksheet[f"{col}{i + 2}"] = team.name
            worksheet[f"{col}{i + 2}"].font = teamNameFont
            worksheet[f"{col}{i + 2}"].fill = PatternFill(patternType="solid", fgColor=TEAM_ROW_COLORS[i])

        # add all stats
        statsWithTitles = self.statSheet(**kwargs).preferredOrderWithTitle()
        for row, teamId in enumerate([team.id for team in self.teams]):
            rowFill = PatternFill(patternType="solid", fgColor=TEAM_ROW_COLORS[row])
            for col, statWithTitle in enumerate(statsWithTitles):
                char = get_column_letter(col + 2)
                if row == 1:
                    # add stat header
                    worksheet[f"{char}{row}"] = statWithTitle[0]
                    worksheet[f"{char}{row}"].font = headerColumnFont
                    worksheet[f"{char}{row}"].fill = headerFill
                # add stat value
                worksheet[f"{char}{row + 2}"] = statWithTitle[1][teamId]
                worksheet[f"{char}{row + 2}"].fill = rowFill

        # put stats into table
        table = Table(displayName=f"YearStats{self.yearNumber}",
                      ref="A1:" + get_column_letter(worksheet.max_column) + str(worksheet.max_row))
        worksheet.add_table(table)

        # save
        workbook.save(filePath)



import os
import tempfile
import unittest

from openpyxl import load_workbook

from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.model.stat.AllTimeStatSheet import AllTimeStatSheet
from test.helper.prototypes import getNDefaultOwnersAndTeams


class TestLeague(unittest.TestCase):
    def test_league_init(self):
        year = Year(yearNumber=0, teams=[], weeks=[])
        owner = Owner(name="")
        league = League(name="leagueName", owners=[owner], years=[year])

        self.assertEqual("leagueName", league.name)
        self.assertEqual(1, len(league.owners))
        self.assertEqual(1, len(league.years))
        self.assertEqual(owner.id, league.owners[0].id)
        self.assertEqual(year.id, league.years[0].id)

    def test_league_statSheet(self):
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        year = Year(yearNumber=2000, teams=teams, weeks=[week])

        league = League(name="TEST", owners=owners, years=[year])

        leagueStatSheet = league.statSheet()

        self.assertIsInstance(leagueStatSheet, AllTimeStatSheet)
        self.assertIsInstance(leagueStatSheet.wins, dict)
        self.assertIsInstance(leagueStatSheet.losses, dict)
        self.assertIsInstance(leagueStatSheet.ties, dict)
        self.assertIsInstance(leagueStatSheet.winPercentage, dict)
        self.assertIsInstance(leagueStatSheet.wal, dict)
        self.assertIsInstance(leagueStatSheet.walPerGame, dict)

        self.assertIsInstance(leagueStatSheet.awal, dict)
        self.assertIsInstance(leagueStatSheet.awalPerGame, dict)
        self.assertIsInstance(leagueStatSheet.opponentAWAL, dict)
        self.assertIsInstance(leagueStatSheet.opponentAWALPerGame, dict)

        self.assertIsInstance(leagueStatSheet.smartWins, dict)
        self.assertIsInstance(leagueStatSheet.smartWinsPerGame, dict)
        self.assertIsInstance(leagueStatSheet.opponentSmartWins, dict)
        self.assertIsInstance(leagueStatSheet.opponentSmartWinsPerGame, dict)

        self.assertIsInstance(leagueStatSheet.pointsScored, dict)
        self.assertIsInstance(leagueStatSheet.pointsScoredPerGame, dict)
        self.assertIsInstance(leagueStatSheet.opponentPointsScored, dict)
        self.assertIsInstance(leagueStatSheet.opponentPointsScoredPerGame, dict)

        self.assertIsInstance(leagueStatSheet.scoringShare, dict)
        self.assertIsInstance(leagueStatSheet.opponentScoringShare, dict)

        self.assertIsInstance(leagueStatSheet.maxScore, dict)
        self.assertIsInstance(leagueStatSheet.minScore, dict)

        self.assertIsInstance(leagueStatSheet.scoringStandardDeviation, dict)

        self.assertIsInstance(leagueStatSheet.plusMinus, dict)

        self.assertIsInstance(leagueStatSheet.teamScore, dict)
        self.assertIsInstance(leagueStatSheet.teamSuccess, dict)
        self.assertIsInstance(leagueStatSheet.teamLuck, dict)

    def test_league_toExcel_happyPath(self):
        owners, teams1 = getNDefaultOwnersAndTeams(2)
        teams1[0].name = "a"
        teams1[1].name = "b"
        matchup1 = Matchup(teamAId=teams1[0].id, teamBId=teams1[1].id, teamAScore=1, teamBScore=2)
        week1 = Week(weekNumber=1, matchups=[matchup1])
        year1 = Year(yearNumber=2000, teams=teams1, weeks=[week1])

        teams2 = [Team(ownerId=owners[0].id, name="a2"), Team(ownerId=owners[1].id, name="b2")]
        matchup2 = Matchup(teamAId=teams2[0].id, teamBId=teams2[1].id, teamAScore=1, teamBScore=2)
        week2 = Week(weekNumber=1, matchups=[matchup2])
        year2 = Year(yearNumber=2001, teams=teams2, weeks=[week2])

        teams3 = [Team(ownerId=owners[0].id, name="a3"), Team(ownerId=owners[1].id, name="b3")]
        matchup3 = Matchup(teamAId=teams3[0].id, teamBId=teams3[1].id, teamAScore=1, teamBScore=2)
        week3 = Week(weekNumber=1, matchups=[matchup3])
        year3 = Year(yearNumber=2002, teams=teams3, weeks=[week3])

        league = League(name="", owners=owners, years=[year1, year2, year3])

        with tempfile.TemporaryDirectory() as tempDir:
            fullPath = os.path.join(tempDir, "tmp.xlsx")
            league.toExcel(fullPath)

            # open created Excel file to check that it was saved correctly
            workbook = load_workbook(filename=fullPath)

            self.assertEqual(4, len(workbook.sheetnames))
            self.assertEqual("2000", workbook.sheetnames[0])
            self.assertEqual("2001", workbook.sheetnames[1])
            self.assertEqual("2002", workbook.sheetnames[2])
            self.assertEqual("All Time", workbook.sheetnames[3])

    def test_league_toExcel_fileAlreadyExists_raisesException(self):
        owners, teams1 = getNDefaultOwnersAndTeams(2)
        teams1[0].name = "a"
        teams1[1].name = "b"
        matchup1 = Matchup(teamAId=teams1[0].id, teamBId=teams1[1].id, teamAScore=1, teamBScore=2)
        week1 = Week(weekNumber=1, matchups=[matchup1])
        year1 = Year(yearNumber=2000, teams=teams1, weeks=[week1])

        teams2 = [Team(ownerId=owners[0].id, name="a2"), Team(ownerId=owners[1].id, name="b2")]
        matchup2 = Matchup(teamAId=teams2[0].id, teamBId=teams2[1].id, teamAScore=1, teamBScore=2)
        week2 = Week(weekNumber=1, matchups=[matchup2])
        year2 = Year(yearNumber=2001, teams=teams2, weeks=[week2])

        teams3 = [Team(ownerId=owners[0].id, name="a3"), Team(ownerId=owners[1].id, name="b3")]
        matchup3 = Matchup(teamAId=teams3[0].id, teamBId=teams3[1].id, teamAScore=1, teamBScore=2)
        week3 = Week(weekNumber=1, matchups=[matchup3])
        year3 = Year(yearNumber=1999, teams=teams3, weeks=[week3])

        league = League(name="", owners=owners, years=[year1, year2, year3])

        with self.assertRaises(FileExistsError) as context:
            with tempfile.TemporaryDirectory() as tempDir:
                fullPath = os.path.join(tempDir, "tmp.xlsx")
                # create a file with this same name/path first
                with open(fullPath, "w") as f:
                    ...
                league.toExcel(fullPath)
        self.assertEqual(f"Cannot create file at path: '{fullPath}' because there is already a file there.",
                         str(context.exception))



from dataclasses import dataclass

from leeger.enum.MatchupType import MatchupType
from leeger.exception.InvalidMatchupFormatException import InvalidMatchupFormatException
from leeger.model.abstract.UniqueId import UniqueId


@dataclass(kw_only=True)
class Matchup(UniqueId):
    teamAId: str
    teamBId: str
    teamAScore: float | int
    teamBScore: float | int
    matchupType: MatchupType = MatchupType.REGULAR_SEASON
    teamAHasTiebreaker: bool = False
    teamBHasTiebreaker: bool = False
    
    def __post_init__(self):
        # Team A and Team B cannot both have the tiebreaker
        if self.teamAHasTiebreaker is True and self.teamBHasTiebreaker is True:
            raise InvalidMatchupFormatException("Team A and Team B cannot both have the tiebreaker.")



from leeger.decorator.validate.validators import validateLeague
from leeger.exception.InvalidFilterException import InvalidFilterException
from leeger.model.filter.StreakFilters import StreakFilters
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.util.LeagueNavigator import LeagueNavigator


class StreakCalculator:
    """
    Should be inherited by all Streak calculators
    """
    MINIMUM_GAMES_FOR_STREAK = 2  # minimum number of games to be considered a streak

    @classmethod
    def _getStreakFilters(cls, league: League, **kwargs) -> StreakFilters:
        onlyChampionship = kwargs.pop("onlyChampionship", False)
        onlyPostSeason = kwargs.pop("onlyPostSeason", False)
        onlyRegularSeason = kwargs.pop("onlyRegularSeason", False)
        yearNumberStart = kwargs.pop("yearNumberStart", league.years[0].yearNumber)
        weekNumberStart = kwargs.pop("weekNumberStart",
                                     LeagueNavigator.getYearByYearNumber(league, yearNumberStart).weeks[0].weekNumber)
        yearNumberEnd = kwargs.pop("yearNumberEnd", league.years[-1].yearNumber)
        weekNumberEnd = kwargs.pop("weekNumberEnd",
                                   LeagueNavigator.getYearByYearNumber(league, yearNumberEnd).weeks[-1].weekNumber)
        onlyOngoing = kwargs.pop("onlyOngoing", False)

        ####################
        # validate filters #
        ####################
        # type checks
        if type(onlyChampionship) != bool:
            raise InvalidFilterException("'onlyChampionship' must be type 'bool'")
        if type(onlyPostSeason) != bool:
            raise InvalidFilterException("'onlyPostSeason' must be type 'bool'")
        if type(onlyRegularSeason) != bool:
            raise InvalidFilterException("'onlyRegularSeason' must be type 'bool'")
        if type(yearNumberStart) != int:
            raise InvalidFilterException("'yearNumberStart' must be type 'int'")
        if type(weekNumberStart) != int:
            raise InvalidFilterException("'weekNumberStart' must be type 'int'")
        if type(yearNumberEnd) != int:
            raise InvalidFilterException("'yearNumberEnd' must be type 'int'")
        if type(weekNumberEnd) != int:
            raise InvalidFilterException("'weekNumberEnd' must be type 'int'")
        if type(onlyOngoing) != bool:
            raise InvalidFilterException("'onlyOngoing' must be type 'bool'")

        # logic checks
        if [onlyChampionship, onlyPostSeason, onlyRegularSeason].count(True) > 1:
            raise InvalidFilterException(
                "Only one of 'onlyChampionship', 'onlyPostSeason', 'onlyRegularSeason' can be True")
        if yearNumberStart > yearNumberEnd:
            raise InvalidFilterException("'yearNumberStart' cannot be greater than 'yearNumberEnd'.")
        if weekNumberStart < 1:
            raise InvalidFilterException("'weekNumberStart' cannot be less than 1.")
        if weekNumberEnd > len(LeagueNavigator.getYearByYearNumber(league, yearNumberEnd).weeks):
            raise InvalidFilterException("'weekNumberEnd' cannot be greater than the number of weeks in the year.")
        if weekNumberStart > weekNumberEnd and yearNumberStart == yearNumberEnd:
            raise InvalidFilterException(
                "'weekNumberStart' cannot be greater than 'weekNumberEnd' within the same year.")

        return StreakFilters(yearNumberStart=yearNumberStart,
                             weekNumberStart=weekNumberStart,
                             yearNumberEnd=yearNumberEnd,
                             weekNumberEnd=weekNumberEnd,
                             onlyChampionship=onlyChampionship,
                             onlyPostSeason=onlyPostSeason,
                             onlyRegularSeason=onlyRegularSeason,
                             onlyOngoing=onlyOngoing)

    @classmethod
    @validateLeague
    def _getAllFilteredMatchupsByWeek(cls, league: League, streakFilters: StreakFilters, **kwargs) -> dict[
        int, list[Matchup]]:
        """
        Returns all Matchups in the given League that are remaining after the given filters are applied.
        Key: Week Number
        Value: List of Matchups
        """

        # parse filters
        yearWeekNumberStartWeekNumberEnd: list[tuple] = list()
        if streakFilters.yearNumberStart == streakFilters.yearNumberEnd:
            yearWeekNumberStartWeekNumberEnd.append(
                (LeagueNavigator.getYearByYearNumber(league, streakFilters.yearNumberStart),
                 streakFilters.weekNumberStart,
                 streakFilters.weekNumberEnd))
        else:
            for year in league.years:
                if year.yearNumber == streakFilters.yearNumberStart:
                    # first year we want, make sure week number start matches what was requested
                    # givenWeekStart and every week greater
                    yearWeekNumberStartWeekNumberEnd.append((year, streakFilters.weekNumberStart, len(year.weeks)))
                elif year.yearNumber == streakFilters.yearNumberEnd:
                    # last year we want, make sure week number end matches what was requested
                    # first week and every week til givenWeekEnd
                    yearWeekNumberStartWeekNumberEnd.append((year, 1, streakFilters.weekNumberEnd))
                elif streakFilters.yearNumberStart < year.yearNumber < streakFilters.yearNumberEnd:
                    # this year is in our year range, include every week in this year
                    yearWeekNumberStartWeekNumberEnd.append((year, 1, len(year.weeks)))

        allFilteredMatchupsByWeek: dict[int, list[Matchup]] = dict()

        for yse in yearWeekNumberStartWeekNumberEnd:
            currentYear = yse[0]
            currentWeekNumberStart = yse[1]
            currentWeekNumberEnd = yse[2]

            for week in currentYear.weeks:
                allFilteredMatchupsByWeek[week.weekNumber] = list()
                if week.weekNumber >= currentWeekNumberStart and week.weekNumber <= currentWeekNumberEnd:
                    for matchup in week.matchups:
                        if matchup.matchupType in streakFilters.includeMatchupTypes:
                            allFilteredMatchupsByWeek[week.weekNumber].append(matchup)

        return allFilteredMatchupsByWeek



import os
import random
from dataclasses import dataclass

from openpyxl import load_workbook
from openpyxl.styles import Font, Color, PatternFill
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.table import Table

from leeger.model.abstract.UniqueId import UniqueId
from leeger.model.league.Owner import Owner
from leeger.model.league.Year import Year
from leeger.model.stat.AllTimeStatSheet import AllTimeStatSheet


@dataclass(kw_only=True)
class League(UniqueId):
    name: str
    owners: list[Owner]
    years: list[Year]

    def statSheet(self, **kwargs) -> AllTimeStatSheet:
        from leeger.calculator.all_time_calculator.AWALAllTimeCalculator import AWALAllTimeCalculator
        from leeger.calculator.all_time_calculator.GameOutcomeAllTimeCalculator import GameOutcomeAllTimeCalculator
        from leeger.calculator.all_time_calculator.PlusMinusAllTimeCalculator import PlusMinusAllTimeCalculator
        from leeger.calculator.all_time_calculator.PointsScoredAllTimeCalculator import \
            PointsScoredAllTimeCalculator
        from leeger.calculator.all_time_calculator.SSLAllTimeCalculator import SSLAllTimeCalculator
        from leeger.calculator.all_time_calculator.ScoringShareAllTimeCalculator import \
            ScoringShareAllTimeCalculator
        from leeger.calculator.all_time_calculator.ScoringStandardDeviationAllTimeCalculator import \
            ScoringStandardDeviationAllTimeCalculator
        from leeger.calculator.all_time_calculator.SingleScoreAllTimeCalculator import SingleScoreAllTimeCalculator
        from leeger.calculator.all_time_calculator.SmartWinsAllTimeCalculator import SmartWinsAllTimeCalculator

        # Game Outcome
        wins = GameOutcomeAllTimeCalculator.getWins(self, **kwargs)
        losses = GameOutcomeAllTimeCalculator.getLosses(self, **kwargs)
        ties = GameOutcomeAllTimeCalculator.getTies(self, **kwargs)
        winPercentage = GameOutcomeAllTimeCalculator.getWinPercentage(self, **kwargs)
        wal = GameOutcomeAllTimeCalculator.getWAL(self, **kwargs)
        walPerGame = GameOutcomeAllTimeCalculator.getWALPerGame(self, **kwargs)

        # AWAL
        awal = AWALAllTimeCalculator.getAWAL(self, **kwargs)
        awalPerGame = AWALAllTimeCalculator.getAWALPerGame(self, **kwargs)
        opponentAWAL = AWALAllTimeCalculator.getOpponentAWAL(self, **kwargs)
        opponentAWALPerGame = AWALAllTimeCalculator.getOpponentAWALPerGame(self, **kwargs)

        # Smart Wins
        smartWins = SmartWinsAllTimeCalculator.getSmartWins(self, **kwargs)
        smartWinsPerGame = SmartWinsAllTimeCalculator.getSmartWinsPerGame(self, **kwargs)
        opponentSmartWins = SmartWinsAllTimeCalculator.getOpponentSmartWins(self, **kwargs)
        opponentSmartWinsPerGame = SmartWinsAllTimeCalculator.getOpponentSmartWinsPerGame(self, **kwargs)

        # Points Scored
        pointsScored = PointsScoredAllTimeCalculator.getPointsScored(self, **kwargs)
        pointsScoredPerGame = PointsScoredAllTimeCalculator.getPointsScoredPerGame(self, **kwargs)
        opponentPointsScored = PointsScoredAllTimeCalculator.getOpponentPointsScored(self, **kwargs)
        opponentPointsScoredPerGame = PointsScoredAllTimeCalculator.getOpponentPointsScoredPerGame(self, **kwargs)

        # Scoring Share
        scoringShare = ScoringShareAllTimeCalculator.getScoringShare(self, **kwargs)
        opponentScoringShare = ScoringShareAllTimeCalculator.getOpponentScoringShare(self, **kwargs)

        # Single Score
        maxScore = SingleScoreAllTimeCalculator.getMaxScore(self, **kwargs)
        minScore = SingleScoreAllTimeCalculator.getMinScore(self, **kwargs)

        # Scoring Standard Deviation
        scoringStandardDeviation = ScoringStandardDeviationAllTimeCalculator.getScoringStandardDeviation(self, **kwargs)

        # Plus Minus
        plusMinus = PlusMinusAllTimeCalculator.getPlusMinus(self, **kwargs)

        # SSL
        teamScore = SSLAllTimeCalculator.getTeamScore(self, **kwargs)
        teamSuccess = SSLAllTimeCalculator.getTeamSuccess(self, **kwargs)
        teamLuck = SSLAllTimeCalculator.getTeamLuck(self, **kwargs)

        return AllTimeStatSheet(wins=wins, losses=losses, ties=ties, winPercentage=winPercentage, wal=wal,
                                walPerGame=walPerGame, awal=awal, awalPerGame=awalPerGame, opponentAWAL=opponentAWAL,
                                opponentAWALPerGame=opponentAWALPerGame, smartWins=smartWins,
                                smartWinsPerGame=smartWinsPerGame, opponentSmartWins=opponentSmartWins,
                                opponentSmartWinsPerGame=opponentSmartWinsPerGame, pointsScored=pointsScored,
                                pointsScoredPerGame=pointsScoredPerGame, opponentPointsScored=opponentPointsScored,
                                opponentPointsScoredPerGame=opponentPointsScoredPerGame, scoringShare=scoringShare,
                                opponentScoringShare=opponentScoringShare, maxScore=maxScore, minScore=minScore,
                                scoringStandardDeviation=scoringStandardDeviation, plusMinus=plusMinus,
                                teamScore=teamScore, teamSuccess=teamSuccess, teamLuck=teamLuck)

    def toExcel(self, filePath: str, **kwargs) -> None:
        if os.path.exists(filePath):
            raise FileExistsError(f"Cannot create file at path: '{filePath}' because there is already a file there.")
        for year in self.years:
            year.toExcel(filePath, **kwargs)

        # add All-Time stats sheet
        workbook = load_workbook(filename=filePath)
        # figure out index to put this sheet into
        # we want the sheets to be ordered: oldest year -> newest year -> all time
        index = len(workbook.sheetnames)
        workbook.create_sheet("All Time", index=index)
        worksheet = workbook["All Time"]

        ####################
        # Styles for table #
        ####################

        # fonts
        headerColumnFont = Font(size=12, bold=True)
        teamNameFont = Font(size=11, bold=True)

        # colors
        GRAY = Color(rgb="B8B8B8")

        def getRandomColor(tint: float = 0) -> Color:
            r = lambda: random.randint(0, 255)
            hexCode = "%02X%02X%02X" % (r(), r(), r())
            return Color(rgb=hexCode, tint=tint)

        OWNER_ROW_COLORS = [getRandomColor(0.5) for _ in range(len(self.owners))]

        # fills
        headerFill = PatternFill(patternType="solid", fgColor=GRAY)

        #################
        # Fill in table #
        #################

        # add title
        worksheet["A1"] = "Owner Names"
        worksheet["A1"].font = headerColumnFont
        worksheet["A1"].fill = headerFill
        # add all team names
        for i, owner in enumerate(self.owners):
            col = "A"
            worksheet[f"{col}{i + 2}"] = owner.name
            worksheet[f"{col}{i + 2}"].font = teamNameFont
            worksheet[f"{col}{i + 2}"].fill = PatternFill(patternType="solid", fgColor=OWNER_ROW_COLORS[i])

        # add all stats
        statsWithTitles = self.statSheet(**kwargs).preferredOrderWithTitle()
        for row, ownerId in enumerate([owner.id for owner in self.owners]):
            rowFill = PatternFill(patternType="solid", fgColor=OWNER_ROW_COLORS[row])
            for col, statWithTitle in enumerate(statsWithTitles):
                char = get_column_letter(col + 2)
                if row == 1:
                    # add stat header
                    worksheet[f"{char}{row}"] = statWithTitle[0]
                    worksheet[f"{char}{row}"].font = headerColumnFont
                    worksheet[f"{char}{row}"].fill = headerFill
                # add stat value
                worksheet[f"{char}{row + 2}"] = statWithTitle[1][ownerId]
                worksheet[f"{char}{row + 2}"].fill = rowFill

        # put stats into table
        table = Table(displayName=f"AllTimeStats",
                      ref="A1:" + get_column_letter(worksheet.max_column) + str(worksheet.max_row))
        worksheet.add_table(table)

        # save
        workbook.save(filePath)



from dataclasses import dataclass


@dataclass(kw_only=True)
class Streak:
    number: int
    teamIdStart: str
    teamIdEnd: str
    weekNumberStart: int
    weekNumberEnd: int
    yearNumberStart: int
    yearNumberEnd: int
    isOngoing: bool



from abc import ABC, abstractmethod

from leeger.model.league.League import League


class LeagueLoader(ABC):
    """
    League Loader classes should inherit this.
    The point of a test_league loader is to load a League object from different Fantasy Football sources.
    """

    @abstractmethod
    def loadLeague(self, *args, **kwargs) -> League:
        ...



from typing import Any


class GeneralUtil:

    @classmethod
    def filter(cls, *, value: Any, list_: list) -> list:
        """
        Filters out the given value from the given list.
        """
        return [v for v in list_ if v is not value]
